import java.util.Vector;
import krister.Ess.AudioChannel;
import processing.core.PApplet;

public class Synth {
	PApplet parent;
	AudioChannel base;// statistics are coming from here
	AudioChannel orig;// the original waveform to imitate

	Synth(PApplet p) {
		parent = p;
		base = new AudioChannel(parent.dataPath("base.wav"));
		orig = new AudioChannel(parent.dataPath("orig.wav"));
	}

	void playResult(float[] result) {
		// plays the resulting sound
		AudioChannel resultCh = new AudioChannel(result.length, 44100);
		resultCh.samples = result;
		resultCh.play();
	}

	void saveResult(float[] result, String filename) {
		// saves the result as a wave file
		AudioChannel resultCh = new AudioChannel(result.length, 44100);
		resultCh.samples = result;
		resultCh.saveSound(parent.dataPath(filename + ".wav"));
	}

	float[] sinusoid(float frequency, int echant) {
		// one complete sinusoidal waveform
		int samplesPerPhase = (int) (echant / frequency);
		float[] sample = new float[samplesPerPhase];
		for (int i = 0; i < samplesPerPhase; i++) {
			sample[i] = (float) Math.sin((double) i * Math.PI * 2
					/ samplesPerPhase);
		}
		return sample;
	}

	float[] imitate() {
		// imitates each waveform of the orig
		// with samples from the base
		Vector<Float> result = new Vector<Float>();
		int cursorB, cursorM, cursorE;// beginning, middle, end points
		cursorB = cursorM = cursorE = 0;
		while (cursorB < orig.samples.length - 1) {
			System.out.println(cursorE + " / " + orig.samples.length);
			// search for the waveform going through 0
			int nb = 4;
			int[] points = getNextCursorPoint(cursorB, nb);
			cursorM = points[(int) (nb / 2) - 1];
			cursorE = points[nb - 1];
			float[] toAdd = getImitationOf(subFloat(orig.samples, cursorB,
					cursorE));
			for (int i = cursorB; i < cursorE; i++) {
				if (i < cursorM && cursorB != 0) {
					float fadeIn = (i - cursorB) / (cursorM - cursorB);
					result.set(i, result.get(i) * (1 - fadeIn)
							+ toAdd[i - cursorB] * fadeIn);
				} else {
					result.add(toAdd[i - cursorB]);
				}
			}
			cursorB = cursorM;
		}
		float[] result2 = new float[result.size()];
		for (int i = 0; i < result.size(); i++) {
			result2[i] = result.get(i);
		}
		return result2;
	}

	private int[] getNextCursorPoint(int cursor, int nb) {
		int[] points = new int[nb];// points to define
		boolean previousSign = getSign(orig.samples[cursor]);
		int encountered = 0;// number of lines through 0
		while (encountered < points.length && cursor < orig.samples.length) {
			if (getSign(orig.samples[cursor]) != previousSign) {
				points[encountered] = cursor;
				encountered++;
				previousSign = getSign(orig.samples[cursor]);
			}
			cursor++;
		}
		if (encountered < points.length) {
			for (int i = encountered; i < points.length; i++) {
				points[i] = orig.samples.length - 1;
			}
		}
		return points;
	}

	private float[] getImitationOf(float[] subFloat) {
		int bestI = 0;// the i that matches best
		float bestDifference = subFloat.length * 2;
		// difference between original and base
		for (int i = 0; i < base.samples.length - subFloat.length; i++) {
			float difference = difference(subFloat, subFloat(base.samples, i, i
					+ subFloat.length), bestDifference);
			if (difference < bestDifference) {
				bestI = i;
				bestDifference = difference;
			}
		}
		float[] result = subFloat(base.samples, bestI, bestI + subFloat.length);
		return result;
	}

	private float difference(float[] subFloat, float[] subFloat2, float max) {
		// returns the difference between two float arrays
		float result = 0;
		for (int i = 0; i < subFloat.length; i++) {
			result += Math.abs(subFloat[i] - subFloat2[i]);
			if (result > max) {
				return result;
			}
		}
		return result;
	}

	private float[] subFloat(float[] samples, int cursorB, int cursorE) {
		float[] result = new float[cursorE - cursorB];
		for (int i = 0; i < result.length; i++) {
			result[i] = samples[cursorB + i];
		}
		return result;
	}

	private boolean getSign(float f) {
		if (f >= 0) {
			return true;
		} else {
			return false;
		}
	}

	float[] generate(int duration, float freq, int oriType) {
		float[] result;
		System.out.println("start the generation");
		float[] ori = new float[0];
		if (oriType == 0)
			ori = sinusoid((float) freq, 44100);
		if (oriType == 1)
			ori = orig.samples;
		if (duration == -1) {
			duration = ori.length;
		}
		result = new float[duration];
		int ordre = 20;
		for (int i = 0; i < duration; i++) {
			System.out.println(i + " / " + duration);
			float[] chunk = extractChunk(ori, i, ordre);
			result[i] = nextSample(chunk, base.samples);
		}
		System.out.println("generation done");
		return result;
	}

	@SuppressWarnings("unused")
	private void drawWavePreview(float[] chunk) {
		for (int i = 0; i < chunk.length; i++) {
			parent.point(i, chunk[i] * 50 + 50);
		}
	}

	private float nextSample(float[] chunk, float[] samples) {
		float lessDiff = chunk.length * 2;
		int lessI = 0;
		for (int i = 0; i < samples.length - chunk.length; i++) {
			float diff = 0;
			for (int i2 = 0; i2 < chunk.length; i2++) {
				float poids = i2;
				diff += Math.abs(samples[i + i2] - chunk[i2]) * poids;
			}
			if (diff < lessDiff) {
				lessDiff = diff;
				lessI = i;
			}
		}
		return samples[lessI + chunk.length - 1];
	}

	private float[] extractChunk(float[] samples, int current, int t) {
		float[] chunk = new float[t];
		for (int i = current - (t - 1); i <= current; i++) {
			int i2 = i;
			while (i2 < 0)
				i2 = i2 + samples.length;
			while (i2 >= samples.length)
				i2 = i2 % samples.length;
			chunk[i - (current - (t - 1))] = samples[i2];
		}
		return chunk;
	}

	public float freqOf(int note) {
		return (float) ((440 / 32) * Math.pow(2, ((float) note - 9) / 12));
	}

	public void recordMidiRange() {
		int c = 1;
		if (c == 0) {
			float[] waveform = generate(-1, 1, 1);
			saveResult(waveform, "result");
		} else {
			for (int i = 0; i < 128; i++) {
				float[] waveform = generate(-1, freqOf(i), 0);
				saveResult(waveform, String.valueOf(i));
			}
		}
	}

	public void saveImitation() {
		saveResult(imitate(), "imitationResult");
	}

}
